=encoding utf8

This file is automatically generated.  DON'T EDIT THIS FILE MANUALLY.

=cut

package Unicode::LineBreak;

use constant {
    M => 2,
    D => 1,
    I => -1,
    P => -2,
};

our @LB_CLASSES = qw(LB_OP LB_CL LB_QU LB_GL LB_NS LB_EX LB_SY LB_IS LB_PR LB_PO LB_NU LB_AL LB_ID LB_IN LB_HY LB_BA LB_BB LB_B2 LB_CB LB_ZW LB_WJ LB_H2 LB_H3 LB_JL LB_JV LB_JT LB_BK LB_CR LB_LF LB_NL LB_SP LB_CM LB_SG LB_AI LB_SA LB_XX);

use constant {
    LB_OP => 0,
    LB_CL => 1,
    LB_QU => 2,
    LB_GL => 3,
    LB_NS => 4,
    LB_EX => 5,
    LB_SY => 6,
    LB_IS => 7,
    LB_PR => 8,
    LB_PO => 9,
    LB_NU => 10,
    LB_AL => 11,
    LB_ID => 12,
    LB_IN => 13,
    LB_HY => 14,
    LB_BA => 15,
    LB_BB => 16,
    LB_B2 => 17,
    LB_CB => 18,
    LB_ZW => 19,
    LB_WJ => 20,
    LB_H2 => 21,
    LB_H3 => 22,
    LB_JL => 23,
    LB_JV => 24,
    LB_JT => 25,
    LB_BK => 26,
    LB_CR => 27,
    LB_LF => 28,
    LB_NL => 29,
    LB_SP => 30,
    LB_CM => 31,
    LB_SG => 32,
    LB_AI => 33,
    LB_SA => 34,
    LB_XX => 35,
    LB_SAal => 36,
    LB_SAcm => 37,
    LB_NSidIter => 38,
    LB_NSidMasu => 39,
    LB_NSidKana => 40,
    LB_NSidLong => 41,
};

use constant {
    EA_z => 0,
    EA_Na => 1,
    EA_N => 2,
    EA_A => 3,
    EA_AnLat => 4,
    EA_AnGre => 5,
    EA_AnCyr => 6,
    EA_W => 7,
    EA_H => 8,
    EA_F => 9,
};

1;
